"""

Configerus Config base.

Here we define the base Config object, which is the typical entrypoint into
Configerus functionality.

"""
import logging
from importlib import metadata
import copy
from typing import Any, Dict

from .plugin import Factory, Type
from .instances import PluginInstances
from .shared import tree_merge
from .loaded import Loaded
from .validator import ValidationError
from .format import Formatter

logger = logging.getLogger("configerus:config")

CONFIGERUS_BOOTSTRAP_ENTRYPOINT = "configerus.bootstrap"
""" SetupTools entrypoint target for bootstrapping """

CONFIG_PATH_LABEL = "paths"
""" If you load this label, it is meant to be return a keyed path """

PLUGIN_DEFAULT_PRIORITY = 75
""" Default plugin priority, which should be a common unprioritized value

General approach:

<35 Low priority defaults (system)
<50 Higher priority defaults (contrib)
<75 Low prority setting (project)
 75 default
<90 High priority settings (project, contrib)
>90 !important (project)

"""


class Config:
    """Config management class (v3).

    Allows some sourced file based project configuration with some pattern
    templating and path set overriding.

    To allow easy separation of config into manageable files, all config is
    grouped into "label" strings which correlate to config files with matching
    names.
    Before accessing a config label, it has to be loaded, which you can do with
    the load() method.
    Then all get() calls use the most recently loaded config label.

    Templating: some string substitution can be done based on config values
        and some path keys. Templating options are based on a custom syntax

    Overriding: tell config about multiple paths in order of priority, and it
        will allow higher priority values to override lower priority values

    Dot notation retrival: you can treat your data as a tree, and config will
        allow you to retrieve tree nodes using a syntax where "." is a node
        label delimiter.
    """

    def __init__(self):
        """Initialize Config object."""
        self.plugins = PluginInstances(self.make_plugin)
        """List of all of the plugins as PluginInstance wrappers

        This mixes plugin types together but but it simplifies management
        and storage of plugins """

        self.loaded = {}
        """ cache of config that has been loaded """

    def copy(self):
        """Make a copy of this config object.

        So that we can independently edit the copy without affecting the
        original.
        """
        config_copy = Config()
        config_copy.plugins = self.plugins.copy(
            config_copy.make_plugin, config_copy.copy_plugin
        )
        return config_copy

    def bootstrap(self, bootstrap_id: str):
        """Make a plugin object instance of a type and key.

        A python module out there will need to have declared an entry_point for
        'mirantis.testing.toolbox.{type}' with entry_point key {name}.
        The entrypoint must be a factory method of signature:
        ```
        def XXXX(conf: configerus.config.Config, instance_id:str) -> {plugin}:
        ```

        An alternative would be if the entrypoint refers to a sub-package here.

        The factory should return the plugin, which is likely going to be some
        kind of an object which has value to the caller of the function. This
        function does not specify what the return needs to be.
        """
        logger.debug(
            "Running configerus bootstrap entrypoint: %s", bootstrap_id
        )
        entry_points = metadata.entry_points()[CONFIGERUS_BOOTSTRAP_ENTRYPOINT]
        for entry_point in entry_points:
            if entry_point.name == bootstrap_id:
                bootstrap_entrypoint = entry_point.load()
                bootstrap_entrypoint(self)
                break
        else:
            raise KeyError(
                "Bootstrap not found {}:{}".format(
                    CONFIGERUS_BOOTSTRAP_ENTRYPOINT, bootstrap_id
                )
            )

    # Pass through accessors
    #
    # These methods just give access to some values used to govern config
    # without having to import the module and directly access constants.

    # pylint: disable=no-self-use
    def paths_label(self):
        """Retrieve the special config label which is a list of paths."""
        return CONFIG_PATH_LABEL

    # pylint: disable=no-self-use
    def default_priority(self):
        """Return the default priority for relative priority setting."""
        return PLUGIN_DEFAULT_PRIORITY

    # Plugin factory

    # This method matches the interface from the instaces object
    # pylint: disable=redefined-builtin, unused-argument
    def make_plugin(
        self,
        type: Type,
        plugin_id: str,
        instance_id: str,
        priority: int = PLUGIN_DEFAULT_PRIORITY,
    ):
        """Make a new plugin."""
        fac = Factory(type, plugin_id)
        return fac.create(self, instance_id)

    def copy_plugin(self, plugin: object):
        """Make a plugin object."""
        if hasattr(plugin, "copy"):
            plugin_copy = plugin.copy()
        else:
            try:
                plugin_copy = copy.deepcopy(plugin)
            except Exception as err:
                raise RuntimeError(f"Could not copy plugin: {plugin}") from err
        plugin_copy.config = self
        return plugin_copy

    # Source plugins
    #
    # All methods related to managing and using source plugins.
    #
    # Some methods are short wrappers on the plugin abstractions, but that
    # allows consumers to not have to import the .plugin.Type enums
    #

    def has_source(self, instance_id: str):
        """Check if a source instance has already been added."""
        return self.plugins.has_plugin(
            instance_id=instance_id, type=Type.SOURCE
        )

    def add_source(
        self,
        plugin_id: str,
        instance_id: str = "",
        priority: int = PLUGIN_DEFAULT_PRIORITY,
    ):
        """Add a new config source to the config object.

        Parameters:
        -----------
        plugin_id (str) : id of the plugin as registered using the plugin
        factory decorater for a source plugin.  This has to match a plugin's
        registration with the plugin factory as a part of the Factory
        decoration

        instance_id (str) : Optionally give a source plugin instance a name,
            which it might use for internal functionality.
            The "path" source plugin allows string template substitution of the
            "__paths__:instance_id" for the path.

        priority (int) : source priority. Use this to set this source values as
            higher or lower priority than others.

        Returns:
        --------
        Returns the source plugin so that you can do any actions to the plugin@
        that it upports, and the code here doesn't need to get fancy with
        @function arguments

        """
        # drop any loaded config
        self.loaded = {}
        # add the plugin to our list.
        return self.plugins.add_plugin(
            Type.SOURCE, plugin_id, instance_id, priority
        )

    def load(
        self, label: str, force_reload: bool = False, validator: str = ""
    ) -> Loaded:
        """Load a config label.

        This queries all sources for the label and merges the loaded config
        data into a single Loaded object.
        The loads are cached per Config instance to make repeated calls cheap.

        Parameters:
        -----------
        label (str) : Config label to load.  Every config source can provide
            config data sets for labels, which correspond to things like
            "all files with a matching filename", or all "records in a matching
            db table"

        force_reload (bool) : if true, and data has been loaded before, this
            forces a fresh reload of data from the sources

        validator (str) : string key passed to the validate() method which will
            validate the retrieved data before returning it.

        Returns:
        --------
        A Loaded object from which you can .get() specific pieces of config

        Throws:
        -------
        Some source plugins will throw Exceptions if they have internal
        problems.

        If you request a validation, and validation fails, then a Validation
        error with be raised.
        """
        if force_reload or label not in self.loaded:
            # Load data from all of the sources for a label
            logger.debug("Loading Config '%s' from all sources", label)

            data: Dict[str, Any] = {}
            # merge in data from the higher priorty into the lower priority
            for source in self.plugins.get_plugins(type=Type.SOURCE):
                source_data = source.load(label)
                if source_data:
                    data = tree_merge(data, source_data)

            if not data:
                raise KeyError(
                    f"Config '{label}' loaded data came out empty.  That means"
                    "that no config source could find that label.  That is "
                    "likely a problem"
                )

            self.loaded[label] = Loaded(
                data=data, parent=self, instance_id=label
            )

        if validator:
            self.validate(self.loaded[label].data, validator)

        logger.debug("Loaded config %s : %s", label, self.loaded[label].data)
        return self.loaded[label]

    # Formatter plugin usage and management

    def has_formatter(self, instance_id: str):
        """Check if a formatter instance has already been added."""
        return self.plugins.has_plugin(
            instance_id=instance_id, type=Type.FORMATTER
        )

    def add_formatter(
        self,
        plugin_id: str,
        instance_id: str = "",
        priority: int = PLUGIN_DEFAULT_PRIORITY,
    ):
        """Add a new config formatter to the config object and return it.

        Parameters:
        -----------
        plugin_id (str) : id of the plugin as registered using the plugin
            factory  decorater for a formatter plugin.  This has to match a
            plugin's registration  with the plugin factory as a part of the
            Factory decoration

        instance_id (str) : Optionally give a formatters plugin instance a
            name, which it might use for internal functionality.

        priority (int) : source priority. Use this to set this source values as
            higher or lower priority than others.

        Returns:
        --------
        Returns the source plugin so that you can do any actions to the plugin
        that it supports, and the code here doesn't need to get fancy with
        function arguments
        """
        return self.plugins.add_plugin(
            Type.FORMATTER, plugin_id, instance_id, priority
        )

    def format(self, data, default_label: Any, validator: str = ""):
        """Format some data using the config object formatters.

        Parameters:
        -----------
        data (Any): primitive data that should be formatted. The data will be
            passed to the formatter plugins in descending priority order.

        default_label (str) : because string templating can indicate a replace
            target without a label source, a default source is used to tell a
            formatter what config label to use in its absence.

        validate_target (Any) : Validation target which will be interpreted by
            the validation plugins.  Two typical formats are common:

            1. a string with some identifiable pattern that a plugin can
                recognize
            2. a dict with keys that different plugins can recognize.

            if empty/None then no validation is performed.

        Raises:
        -------
        Will throw a ValueError if your data contains a format tag that cannot
        be interpreted.
        Will raise a KeyError if your data contains a format tag with a bad key
        for action, or for a value as interpreted by the plugin.
        """
        formatter = Formatter(self)
        data = formatter.format(data=data, default_label=default_label)

        if validator:
            self.validate(data, validator)

        return data

    # Validator plugin usage and management

    def has_validator(self, instance_id: str):
        """Check if a formatter instance has already been added."""
        return self.plugins.has_plugin(
            instance_id=instance_id, type=Type.VALIDATOR
        )

    def add_validator(
        self,
        plugin_id: str,
        instance_id: str = "",
        priority: int = PLUGIN_DEFAULT_PRIORITY,
    ):
        """Add a new config validator to the config object and return it.

        Parameters:
        -----------
        plugin_id (str) : id of the plugin as registered using the plugin
            factory  decorater for a validator plugin.  This has to match a
            plugin's registration  with the plugin factory as a part of the
            Factory decoration

        instance_id (str) : Optionally give a validator plugin instance a name,
            which it might use for internal functionality.

        priority (int) : source priority. Use this to set this source values as
            higher or lower priority than others.

        Returns:
        --------
        Returns the source plugin so that you can do any actions to the plugin
        that it supports, and the code here doesn't need to get fancy with
        function arguments
        """
        return self.plugins.add_plugin(
            Type.VALIDATOR, plugin_id, instance_id, priority
        )

    def validate(
        self, data, validate_target: str, exception_if_invalid: bool = True
    ):
        """Format some data using the config object formatters.

        data (Any): primitive data that should be formatted. The data will be
            passed to the formatter plugins in descending priority order.

        validate_target : A config key which can be used

        Returns:
        --------
        Bool results of validation. If any validation plugin raises an
        exception then it will be false.

        If exception_if_invalid then validation failures with raise and@
        exception

        Raises:
        -------
        ValidationError on any validation exception.

        ValdiationError if any validator ran into an exception internally.

        @TODO can we be more specific in interpreting validation exceptions.
        """
        try:
            for validator in self.plugins.get_plugins(
                type=Type.VALIDATOR, exception_if_missing=False
            ):
                validator.validate(validate_target, data)
        # pylint: disable=broad-except
        except Exception as err:
            if exception_if_invalid:
                raise ValidationError(
                    "Config validation failed: {err}"
                ) from err
            return False
        return True



"""

Configerus contrib package for config from environment variables.

"""
import logging

from configerus.config import Config
from configerus.plugin import SourceFactory

from .env_specific_source import (
    ConfigSourceEnvSpecificPlugin,
    PLUGIN_ID_SOURCE_ENV_SPECIFIC,
    CONFIGERUS_ENV_SPECIFIC_BASE_KEY,
)
from .env_json_source import (
    ConfigSourceEnvJsonPlugin,
    PLUGIN_ID_SOURCE_ENV_JSON,
    CONFIGERUS_ENV_JSON_ENV_KEY,
)


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_ENV_SPECIFIC)
def plugin_factory_configsource_env_specific(
    config: Config, instance_id: str = ""
):
    """Create an configsource specific env plugin."""
    return ConfigSourceEnvSpecificPlugin(config, instance_id)


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_ENV_JSON)
def plugin_factory_configsource_env_json(
    config: Config, instance_id: str = ""
):
    """Create an configsource json env plugin."""
    return ConfigSourceEnvJsonPlugin(config, instance_id)


# Unused config arg is a part of the bootstrap interface.
# pylint: disable=unused-argument
def configerus_bootstrap(config: Config):
    """Bootstrap a config object.

    We don't actually do anything, so this bootstrapper is here only to ensure
    that the above factory decorators are run

    """



"""

test the plugin instance functionality

"""
import logging
import unittest

from configerus.config import Config
from configerus.plugin import (
    SourceFactory,
    FormatFactory,
    ValidatorFactory,
    Type,
)
from configerus.instances import PluginInstance, PluginInstances

logger = logging.getLogger("test_instances")

""" register a bunch of dummy plugins which do nothing but can be used for testing """


@SourceFactory(plugin_id="dummy_1")
def plugin_factory_source_1(config: Config, instance_id: str = ""):
    """create a dummy configsource plugin"""
    return DummyPlugin(config, instance_id)


@SourceFactory(plugin_id="dummy_2")
def plugin_factory_source_2(config: Config, instance_id: str = ""):
    """create a dummy configsource plugin"""
    return DummyPlugin(config, instance_id)


@FormatFactory(plugin_id="dummy_1")
def plugin_factory_format_1(config: Config, instance_id: str = ""):
    """create a dummy formatter plugin"""
    return DummyPlugin(config, instance_id)


@FormatFactory(plugin_id="dummy_2")
def plugin_factory_format_2(config: Config, instance_id: str = ""):
    """create a dummy formatter plugin"""
    return DummyPlugin(config, instance_id)


@ValidatorFactory(plugin_id="dummy_1")
def plugin_factory_validate_1(config: Config, instance_id: str = ""):
    """create a dummy formatter plugin"""
    return DummyPlugin(config, instance_id)


@ValidatorFactory(plugin_id="dummy_2")
def plugin_factory_validate_2(config: Config, instance_id: str = ""):
    """create a dummy formatter plugin"""
    return DummyPlugin(config, instance_id)


class DummyPlugin:
    """Just a placehold plugin object"""

    def __init__(self, config: Config, instance_id: str):
        self.config = config
        self.instance_id = instance_id


class ConfigTemplating(unittest.TestCase):
    def test_instance_struct_sanity(self):
        """just test that the Instance struct has the properties that we use"""
        config = Config()

        instance = PluginInstance(
            plugin_id="dummy",
            instance_id="dummy",
            type=Type.SOURCE,
            priority=60,
            plugin=DummyPlugin(config, "dummy"),
        )

        self.assertEqual(instance.plugin_id, "dummy")
        self.assertEqual(instance.priority, 60)

    def test_instances_sanity(self):
        """Test some construction sanity on the instances object"""
        config = Config()
        instances = PluginInstances(config.make_plugin)

        self.assertEqual(len(instances), 0)

        instances.add_plugin(
            Type.SOURCE,
            "dummy_1",
            "dummy_instance_1",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_1",
            "dummy_instance_2",
            config.default_priority(),
        )

        self.assertEqual(len(instances), 2)

    def test_instances_plugin_get_simple(self):
        """test that we can add plugins and then retrieve them"""
        config = Config()
        instances = PluginInstances(config.make_plugin)

        instances.add_plugin(
            Type.SOURCE,
            "dummy_1",
            "dummy_instance_1",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_1",
            "dummy_instance_2",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_2",
            "dummy_instance_3",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_2",
            "dummy_instance_4",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_1",
            "dummy_instance_5",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_2",
            "dummy_instance_6",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.SOURCE,
            "dummy_1",
            "dummy_instance_7",
            config.default_priority(),
        )

        len(instances) == 7

        get_4 = instances["dummy_instance_4"]
        self.assertTrue(get_4)
        self.assertEqual(get_4.instance_id, "dummy_instance_4")
        self.assertEqual(
            instances.get_plugin(instance_id="dummy_instance_4").instance_id,
            "dummy_instance_4",
        )
        with self.assertRaises(KeyError):
            instances.get_plugin(instance_id="no_such_instance")

    def test_instances_get_plugins(self):
        """get plugins based on multiple search parameters"""
        config = Config()
        instances = PluginInstances(config.make_plugin)

        instances.add_plugin(
            Type.SOURCE, "dummy_1", "dummy_source_1", config.default_priority()
        )
        instances.add_plugin(
            Type.SOURCE, "dummy_1", "dummy_source_2", config.default_priority()
        )
        instances.add_plugin(
            Type.SOURCE, "dummy_2", "dummy_instance", config.default_priority()
        )
        instances.add_plugin(
            Type.FORMATTER,
            "dummy_2",
            "dummy_formatter_4",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.FORMATTER,
            "dummy_1",
            "dummy_formatter_5",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.FORMATTER,
            "dummy_2",
            "dummy_formatter_6",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.VALIDATOR,
            "dummy_1",
            "dummy_instance",
            config.default_priority(),
        )
        instances.add_plugin(
            Type.VALIDATOR,
            "dummy_2",
            "dummy_validator_8",
            config.default_priority(),
        )

        self.assertTrue(instances.has_plugin(instance_id="dummy_formatter_4"))
        self.assertTrue(instances.has_plugin(instance_id="dummy_validator_8"))
        self.assertTrue(instances.has_plugin(instance_id="dummy_formatter_4"))
        self.assertTrue(instances.has_plugin(instance_id="dummy_source_2"))

        self.assertEqual(len(instances.get_instances(type=Type.SOURCE)), 3)
        self.assertEqual(len(instances.get_instances(type=Type.FORMATTER)), 3)
        self.assertEqual(len(instances.get_instances(type=Type.VALIDATOR)), 2)

        self.assertEqual(
            len(
                instances.get_instances(type=Type.SOURCE, plugin_id="dummy_1")
            ),
            2,
        )
        self.assertEqual(
            len(
                instances.get_instances(
                    type=Type.VALIDATOR, plugin_id="dummy_1"
                )
            ),
            1,
        )

        # Plugin ID only search should cross type
        self.assertEqual(len(instances.get_instances(plugin_id="dummy_1")), 4)
        # There are no rules against repeated instance_ids
        self.assertEqual(
            len(instances.get_instances(instance_id="dummy_instance")), 2
        )
        # no rule against empty filtering (gets sorted instances)
        self.assertEqual(len(instances.get_instances()), len(instances))

    def test_instance_priority_simple(self):
        """test that retrieving instances sorts"""
        config = Config()
        instances = PluginInstances(config.make_plugin)

        starting_range = range(30, 71, 10)
        priority_list = range(70, 29, -10)

        for priority in starting_range:
            instances.add_plugin(
                Type.SOURCE,
                "dummy_1",
                "instance_{}".format(priority),
                priority,
            )

        instance_list = instances.get_instances(type=Type.SOURCE)
        self.assertEqual(len(instance_list), len(starting_range))

        for index, priority in enumerate(priority_list):
            self.assertEqual(instance_list[index].priority, priority)

        # let's see what happens as things change
        instances.add_plugin(
            Type.SOURCE, "dummy_1", "instance_{}".format(90), 90
        )
        instance_list = instances.get_instances(type=Type.SOURCE)

        self.assertEqual(len(instance_list), len(starting_range) + 1)
        self.assertEqual(instance_list[0].priority, 90)
        self.assertEqual(instance_list[1].priority, 70)
        self.assertEqual(instance_list[len(starting_range)].priority, 30)



"""

Configerus contrib package for config from passed Dicts.

"""
import logging

from configerus.config import Config
from configerus.plugin import SourceFactory

from .source import ConfigSourceDictPlugin

PLUGIN_ID_SOURCE_DICT = "dict"
"""ConfigSource plugin_id for the configerus dict configsource plugin."""

CONFIGERUS_DICT_DATA_KEY = "data"
"""Key that should give the path."""


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_DICT)
def plugin_factory_configsource_dict(config: Config, instance_id: str = ""):
    """Create an configsource dict plugin."""
    return ConfigSourceDictPlugin(config, instance_id)


# Unused config arg is a part of the bootstrap interface.
# pylint: disable=unused-argument
def configerus_bootstrap(config: Config):
    """Bootstrap a config object.

    We don't actually do anything, so this bootstrapper is here only to ensure
    that the above factory decorator is run

    """



"""

Configerus source plugin that retrieves from a passed Dict.

"""
from typing import Dict, Any

import copy

from configerus.config import Config


class ConfigSourceDictPlugin:
    """Config source plugin that looks for data in a passed Dict."""

    def __init__(self, config: Config, instance_id: str):
        """Initialize the plugin."""
        self.config: Config = config
        self.instance_id: str = instance_id

        self.data: Dict[str, Any] = {}
        """Data that we will use for searching."""

    def copy(self):
        """Make a copy of this plugin."""
        plugin_copy = ConfigSourceDictPlugin(self.config, self.instance_id)
        plugin_copy.set_data(copy.copy(self.data))
        return plugin_copy

    def set_data(self, data: Dict[str, Any]):
        """Assign Dict data to this config source plugin."""
        self.data = data

    def load(self, label: str) -> Dict[str, Any]:
        """Load a config label and return a Dict[str, Any] of config data.

        Parameters:
        -----------
        label (str) : label to load
        """
        return self.data[label] if label in self.data else {}



"""

Configerus contrib package for config from files.

"""
from configerus.config import Config
from configerus.plugin import FormatFactory, SourceFactory

from .source import ConfigSourcePathPlugin
from .format import ConfigFormatFilePlugin


CONFIGERUS_PATH_KEY = "path"
"""Default key that for the path."""

PLUGIN_ID_SOURCE_PATH = "path"
"""ConfigSource plugin_id for the configerus path configsource plugin."""


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_PATH)
def plugin_factory_configsource_path(config: Config, instance_id: str = ""):
    """Create an configsource path plugin."""
    return ConfigSourcePathPlugin(config, instance_id)


PLUGIN_ID_FORMAT_FILE = "file"
"""Format plugin_id for the configerus filepath format plugin."""


@FormatFactory(plugin_id=PLUGIN_ID_FORMAT_FILE)
def plugin_factory_format_file(config: Config, instance_id: str = ""):
    """Create an format plugin which replaces from file contents."""
    return ConfigFormatFilePlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """Bootstrap a config object."""
    config.add_formatter(
        plugin_id=PLUGIN_ID_FORMAT_FILE,
        instance_id=PLUGIN_ID_FORMAT_FILE,
        priority=40,
    )



"""

Configerus format plugins which can replace markers with file contents.

Replacing can be with flat file contents or with parsed json or yml.

"""
import os.path
import re
import json
import logging

import yaml

from configerus.config import Config

FILES_FORMAT_MATCH_PATTERN = r"(?P<file>(\~?\/?\w+\/)*\w*(\.\w+)?)"
""" A regex pattern to identify files that should be embedded """

logger = logging.getLogger("configerus.contrib.files:format")


class ConfigFormatFilePlugin:
    """Format a key by returning the contents of a file."""

    def __init__(self, config: Config, instance_id: str):
        """Initialize the plugin."""
        self.config = config
        self.instance_id = instance_id

        self.pattern = re.compile(FILES_FORMAT_MATCH_PATTERN)

    def copy(self):
        """Make a copy of this plugin."""
        plugin_copy = ConfigFormatFilePlugin(self.config, self.instance_id)
        return plugin_copy

    # pylint: disable=unused-argument
    def format(self, key, default_label: str):
        """Format a string by substituting config values.

        Parameters
        ----------
        key: a string that should gies instructions to the formatter on how to
            create a format replacements

        default_label : if format/replace patterns don't have a source defined
            then this is used as a source.

        Raises
        ------
        FileNotFound is a file replacement is requested but the file path
            cannot be opened

        json.decoder.JSONDecodeError if a json file cannot be unmarshalled

        yaml.YAMLError if a yaml file cannot be unmarshalled

        Returns
        -------
        unmarshalled json/yml file or string contents of the file
        """
        match = self.pattern.fullmatch(key.strip())
        if not match:
            raise KeyError(
                "Could not interpret Format action target '{}'".format(key)
            )

        # path to the file to return as a replacement
        file = match.group("file")
        # file extension, used to make decisions about parsing/unmarshallin
        extension = ""
        file_split = os.path.splitext(file)
        if len(file_split) > 0:
            extension = file_split[1].lower()

        try:
            with open(file) as file_object:
                if extension == ".json":
                    try:
                        return json.load(file_object)
                    except json.decoder.JSONDecodeError as err:
                        raise ValueError(
                            f"Failed to parse one of the config files '{file}'"
                        ) from err

                elif extension in [".yml", ".yaml"]:
                    try:
                        return yaml.safe_load(file_object)
                    except yaml.YAMLError as err:
                        raise ValueError(
                            f"Failed to parse one of the config files '{file}'"
                        ) from err

                # return file contents as a string (above parsing didn't happen)
                return file_object.read()

        except FileNotFoundError as err:
            raise KeyError(
                f"Could not embed file as config as file could not be found: {file}"
            ) from err



"""

Configerus source plugins which load config from file contents.

"""
import os
import re
import logging
from typing import Dict, Any
import json
import copy

import yaml

from configerus.config import Config
from configerus.shared import tree_merge

logger = logging.getLogger("configerus.contrib.source")

# FileTypes that this class can use at this time
FILESOURCE_FILETYPES = ["json", "yaml", "yml"]
""" Valid config types that the config loader can currently handled """

CONFIGERUS_PATH_LABEL = "paths"
""" If you load this label, it is meant to be return a keyed path """


class ConfigSourcePathPlugin:
    """Configerus source plugin that reads files."""

    def __init__(self, config: Config, instance_id: str):
        """Initialize the plugin."""
        self.config = config
        self.instance_id = instance_id

        self.path = ""

    def copy(self):
        """Make a copy of this plugin."""
        plugin_copy = ConfigSourcePathPlugin(self.config, self.instance_id)
        plugin_copy.set_path(copy.deepcopy(self.path))
        return plugin_copy

    def set_path(self, path: str):
        """Set the config path source."""
        self.path = path

    def load(self, label: str):
        """Load config for a name.

        Parameters:
        -----------
        lable (str) : config label to load, should correlated to a json or yaml
            file of the same name in the path, otherwise an empty Dict is
            returned.

            **There is 1 special case, where if MTT_CONFIG_PATH_LABEL is passed
              then the function returns a Dict of 'instance_id:path' which can
              be used for string substitution**

        Returns:
        --------
        Dict[str, Any] of data that was loaded for the label
        """
        if not os.path.isdir(self.path):
            raise ValueError(
                "Could not load '{}' path config, as the source path does not exist: {}".format(
                    self.instance_id, self.path
                )
            )

        # Special case for retreiving paths instead of config
        if label == CONFIGERUS_PATH_LABEL:
            return {self.instance_id: self.path}

        # hold all merged data from found source files
        data: Dict[str, Any] = {}
        # regex part for allowed file extensions
        file_types_re = "|".join(FILESOURCE_FILETYPES)
        # regex that matches all valid config filenames for the label
        config_files_re = rf"({label})\.({file_types_re})"

        for file in [
            f for f in os.listdir(self.path) if re.match(config_files_re, f)
        ]:
            with open(os.path.join(self.path, file)) as matching_file:
                extension = os.path.splitext(file)[1].lower()
                if extension == ".json":
                    try:
                        file_config = json.load(matching_file)
                    except json.decoder.JSONDecodeError as err:
                        raise ValueError(
                            f"Failed to parse one of the config files '{file}'"
                        ) from err

                    assert file_config, f"Empty config in {file} from file."
                elif extension in [".yml", ".yaml"]:
                    try:
                        file_config = yaml.load(
                            matching_file, Loader=yaml.FullLoader
                        )
                    except yaml.YAMLError as err:
                        raise ValueError(
                            f"Failed to parse one of the config files '{file}'"
                        ) from err

                    assert file_config, f"Empty config in {file} [{self.path}]"

                else:
                    raise ValueError(
                        f"Unknown config filetype. Cannot parse '{extension}'"
                        " files, but it matches our regex"
                    )

                data = tree_merge(file_config, data)

        return data



"""

Test config validation

Here we run some simple config validation on both load() and get() operations
using the configerus.contric.jsonscheme validator plugin.

"""

from configerus.contrib.jsonschema import PLUGIN_ID_VALIDATE_JSONSCHEMA
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
import configerus
import unittest
import logging

logging.basicConfig()
logger = logging.getLogger("validate")
logger.setLevel(level=logging.INFO)


""" TESTS """


class JsonSchemaValidate(unittest.TestCase):
    """Some simple validation testing usinf jsonschema"""

    # examples from https://python-jsonschema.readthedocs.io/en/stable/
    valid_instance = {"name": "Eggs", "price": 34.99}
    invalid_instance = {"name": "Eggs", "price": "Invalid"}
    instance_schema = {
        "type": "object",
        "properties": {
            "price": {"type": "number"},
            "name": {"type": "string"},
        },
    }

    def _simple_validate_config(self):
        """use a common config across some tests"""

        logger.debug("Building jsonschema config object")
        config = configerus.new_config()
        config.add_validator(PLUGIN_ID_VALIDATE_JSONSCHEMA)
        config.add_source(PLUGIN_ID_SOURCE_DICT).set_data(
            {
                "valid_load_test": self.valid_instance,
                "invalid_load_test": self.invalid_instance,
                "get_test": {
                    "valid": self.valid_instance,
                    "1": {"invalid": self.invalid_instance},
                },
                "jsonschema": {"instance": self.instance_schema},
            }
        )

        return config

    def test_validate_jsonschema_load_valid(self):
        """just do a minimal jsonschema validation"""

        config = self._simple_validate_config()
        load_config = config.load(
            "valid_load_test", validator="jsonschema:instance"
        )

    def test_validate_jsonschema_load_invalid(self):
        """just do a minimal jsonschema validation"""

        config = self._simple_validate_config()
        with self.assertRaises(Exception):
            config.load("invalid_load_test", validator="jsonschema:instance")

    def test_validate_jsonschema_get(self):
        """just do a minimal jsonschema validation"""

        config = self._simple_validate_config()
        test_config = config.load("get_test")

        # this should be valid
        valid_instance = test_config.get(
            "valid", validator="jsonschema:instance"
        )

        with self.assertRaises(Exception):
            test_config.get("1.invalid", validator="jsonschema:instance")



"""

Configuration validation.

Shared Validation code, which really just comes down to the shared Exception
used to inidcate a validation failure.

"""


class ValidationError(ValueError):
    """Configerus Validation has failed."""



"""

Config source from JSON contents of an ENV variable

This plugin allows you to include JSON in a single ENV variable which if
present will be used as a config source.

This allows large runtime config overrides with out a lot of individual ENV
variables being needed.

If you find that your JSON is tiny and difficult to design/manage, then
confider usng the 'specific' env source plugin.

string case for keys is a weird topic.  We .lower() all keys as it seems a
decent convention.

"""
from typing import Dict, Any
import logging
import os

import copy

from configerus.config import Config

logger = logging.getLogger("configerus.contrib.env.source.specific")

PLUGIN_ID_SOURCE_ENV_SPECIFIC = "env-specific"
""" ConfigSource plugin_id for the configerus env configsource plugin """

CONFIGERUS_ENV_SPECIFIC_BASE_KEY = "base"
""" Config key for retreiving the env specific base value from config """


class ConfigSourceEnvSpecificPlugin:
    """Get config from all ENV variables."""

    def __init__(self, config: Config, instance_id: str):
        """Initialize the plugin."""
        self.config: Config = config
        self.instance_id: str = instance_id

        self.base: str = ""
        """ keep the ENV base prefix that should limit what gets pulled in """

    def copy(self):
        """Make a copy of this plugin."""
        plugin_copy = ConfigSourceEnvSpecificPlugin(
            self.config, self.instance_id
        )
        plugin_copy.set_base(copy.copy(self.base))
        return plugin_copy

    def set_base(self, base: str):
        """Set the ENV base to make the plugin not interpret all ENV vars."""
        self.base = base

    def load(self, label: str) -> Dict[str, Any]:
        """Load a config label and return a Dict[str, Any] of config data.

        Parameters:
        -----------
        label (str) : label to load
        """
        if self.base:
            label_prefix = "{}_{}_".format(self.base, label).upper()
        else:
            label_prefix = "{}_".format(label).upper()

        filtered = {}
        for env_key, env_value in os.environ.items():
            if env_key.upper().startswith(label_prefix):
                filtered[env_key[len(label_prefix) :].lower()] = env_value

        organized: Dict[str, Any] = {}
        for (key, value) in filtered.items():
            if key.startswith("_"):
                continue

            base = organized
            key_steps = key.lower().split("_")

            last_step = key_steps.pop()

            for key_step in key_steps:
                if not isinstance(base, dict):
                    base = {}
                if key_step not in base:
                    base[key_step] = {}
                base = base[key_step]
            if not isinstance(base, dict):
                base = {}
            base[last_step] = value

        return organized



"""

Configerus validation plugin that uses jsonschema to validate.

"""
from typing import Any
from jsonschema import validate

from configerus.config import Config

PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL = "jsonschema"


class JsonSchemaValidatorPlugin:
    """Configerus validation plugin that uses jsonschema to validate."""

    def __init__(self, config: Config, instance_id: str):
        """Initialize plugin."""
        self.config = config
        self.instance_id = instance_id

    def copy(self):
        """Make a copy of this plugin."""
        plugin_copy = JsonSchemaValidatorPlugin(self.config, self.instance_id)
        return plugin_copy

    def validate(self, validate_target: Any, data):
        """Validate a structure using jsonschema pulled from a config key.

        Parameters:
        -----------
        validate_target (str|dict[str:dict]) : validation schema indicatior:

            One of two options will be validated, otherwise we will ignore:

            1. A string of the format "jsonschema:{validation-config-key}"
            2. dict with key "jsonschema"

            For #1, the validation target is expected to be a config target
            (label:key) which will retrieve jsonschema schema from config

            For #2 a dict value for key 'jsonschema' is expected to be the
            schema.

        data (Any) : data which should be validated

        Raises:
        -------
        If a jsonchema schema was identified in the validate target, then a
        jsonschema validation error will be raised if the data is not valid.
        """
        if isinstance(validate_target, str):
            (method, validate_key) = validate_target.split(":")

            if not method == PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL:
                # returns of any value signal validation as validate only
                # catches exceptions
                return

            # This case means that we were told to look for a jsonschema source
            # in config, which we can find by loading jsonschema as a config
            # label and then treating the passed string as a config key for
            # .get()

            try:
                schema_config = self.config.load(
                    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
                )
                schema = schema_config.get(validate_key)

            except Exception as err:
                raise NotImplementedError(
                    "Could not access jsonschema validation schema from config"
                    f" target "
                    f"'{PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL}:"
                    f"{validate_key}'"
                ) from err

        elif isinstance(validate_target, dict):
            # in this case, the validate target is itself a Dict JsonSchema
            # validation definition.

            if (
                PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
                not in validate_target
            ):
                return

            schema = validate_target[
                PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
            ]

            if not isinstance(schema, dict):
                raise ValueError(
                    f"JSONSCHEMA scheme was expected to be a dict: {schema}"
                )

        else:
            # Could not interpret validate target
            return

        # Call the jsonschema validation using the schema.
        # this will raise an exception on validatio failure
        # pylint: disable=not-callable
        validate(data, schema=schema)



"""

Configerus format plugins which can replace markers with other config contents.

"""
import re
import logging

from configerus.config import Config

logger = logging.getLogger("configerus.contrib.get:formatter")

CONFIG_DEFAULT_MATCH_PATTERN = r"((?P<label>\w+)\:)?(?P<key>[-\w.]+)"
""" Default regex pattern used detect format targets for config retrieval """


class ConfigFormatGetPlugin:
    """Return a format replacement by retriving data from config."""

    def __init__(self, config: Config, instance_id: str):
        """Initialize the plugin."""
        self.config = config
        self.instance_id = instance_id

        self.pattern = re.compile(CONFIG_DEFAULT_MATCH_PATTERN)

    def copy(self):
        """Make a copy of this plugin."""
        plugin_copy = ConfigFormatGetPlugin(self.config, self.instance_id)
        return plugin_copy

    def format(self, key, default_label: str):
        """Format a key by returning config values.

        Parameters:
        -----------
        key: a string that should gies instructions to the formatter on how to
            create a format replacements

        default_label : if format/replace patterns don't have a label defined
            then this is used as a label.

        Raises:
        -------
        KeyError if the source or the key could not be found.
        """
        match = self.pattern.fullmatch(key.strip())
        if not match:
            raise KeyError(
                "Could not interpret Format action key '{}'".format(key)
            )

        # from an re.match get data from config.@
        label = match.group("label")
        key = match.group("key")

        if label is None:
            label = default_label

        if label is None:
            label = default_label

        loaded = self.config.load(label)
        return loaded.get(key)



"""

Configerus base package.

Contains typical config constructor for starting to use configerus.

"""
import logging
from typing import List

from .config import Config

logger = logging.getLogger("configerus")


CONFIGERUS_BOOSTRAP_DEFAULT = ["get", "files"]
""" Default list of modules to bootstrap for a new config object """


def new_config(bootstraps: List[str] = None):
    """Get a new Config object.

    bootstraps (List[str]) : list of modules to bootstrap for this config
        object For each string, config will try to run a setuptools entrypoint
        bootstrap function, passing in the config object.

        This can be used for 2 purposes:
        1. the bootstrap function may be in a file which registers plugins that
            can be used in the config object
        2. the bootstrap can modify the config object as needed.

    """
    # start with a new config object
    config = Config()

    if bootstraps is None:
        bootstraps = CONFIGERUS_BOOSTRAP_DEFAULT

    for bootstrap_entrypoint_id in bootstraps:
        logger.debug("bootstrapping: %s", bootstrap_entrypoint_id)
        config.bootstrap(bootstrap_entrypoint_id)

    return config



"""

Common shared functions.

The important methods here are the functions used for navigating trees of Dict
data using the dot notation syntax and the merging of those trees.

"""
import logging
from typing import Dict, Any, List

logger = logging.getLogger("configerus.shared")

# @see https://stackoverflow.com/
#       @questions/20656135/python-deep-merge-dictionary-data


def tree_merge(source: Any, destination: Any):
    """Deep merge source into destination.

    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> merge(b, a) ==
         { 'first' : { 'all_rows' :
           { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
    True
    """
    if not (isinstance(source, dict) and isinstance(destination, dict)):
        return source

    for key, value in source.items():
        if isinstance(value, dict) and key in destination:
            value = tree_merge(value, destination[key])
        destination[key] = value

    return destination


def tree_get(
    node: Dict, keys: List[str], glue: str = ".", ignore: List[str] = None
) -> Any:
    """Find a path down a tree using the keys as a step by step path."""
    if not node:
        raise ValueError(
            "There was no data in the config so no key match could be made"
        )

    flat_steps = tree_reduce(tree=keys, glue=glue, ignore=ignore)
    if len(flat_steps) == 0:
        return node

    for step in flat_steps:
        try:
            if node is None:
                raise KeyError("Path tried to descend into None")
            if isinstance(node, str):
                raise KeyError(
                    "Path tried to descend into a string: {}".format(node)
                )
            if isinstance(node, list) and step.isnumeric():
                node = node[int(step)]
            else:
                # hopefully the target is subscriptable?
                node = node[step]

        except KeyError as err:
            raise KeyError(
                f"Key {keys} not found in loaded config data. '{step}' "
                "was not found"
            ) from err
        except IndexError as err:
            raise IndexError(
                f"Array index '{step}' was not found in list : {node}"
            ) from err
        except TypeError as err:
            raise ValueError(
                f"Invalid key '{step}' in the keys list: '{node}'"
            ) from err

    return node


def tree_reduce(
    tree: Any, glue: str = ".", ignore: List[Any] = None
) -> List[str]:
    """Merge a nested tree of strings down to a flat list of strings.

    Also split any strings that contain a glue character into a list of strings
    and flatten them down as well.

    Also ignore certain passed in strings from the list.

    Parameters:
    -----------
    tree (str | List[str] | List[List[str]] ...) : Any list tree of strings
        that need to be flattened to a single depth

    glue (str) : if a string is found that has a glue, then it is split along
        that glue and converted to a List

    ignore (List[str]) : a list of strings that should be considered invalid
        values that should not be included in the flattenned list

    Returns:
    --------
    List[str] tree reduced to a flat (single) depth, potentially empty
    """
    # don't use the iterator solution with strings
    if isinstance(tree, str):
        if glue == "":
            tree = [tree]
        else:
            tree = tree.split(glue)
    elif isinstance(tree, int):
        tree = [tree]

    else:
        flatter = []
        for node in tree:
            flat_node = tree_reduce(node, glue, ignore)
            flatter += flat_node
        tree = flatter

    if ignore is None:
        ignore = []
    return [node for node in tree if node and node not in ignore]



"""

Test .sharted functions

Simple tests that ensure that the .shared functionality works

"""

import unittest
import logging

import configerus.shared as shared

logger = logging.getLogger("shared")


class BasicConstruct(unittest.TestCase):
    def test_1_tree_reduce(self):
        """can we reduce trees to lists"""
        tree = ["1", "2", "3"]
        self.assertEqual(shared.tree_reduce(tree), tree)

        tree = ["1", "2", ["3", "4"]]
        flat = ["1", "2", "3", "4"]
        self.assertEqual(shared.tree_reduce(tree, glue="."), flat)

        tree = ["1", "2", ["3", "4"], [["5"], "6", "7", [[["8"]]]]]
        flat = ["1", "2", "3", "4", "5", "6", "7", "8"]
        self.assertEqual(shared.tree_reduce(tree, glue="."), flat)

    def test_2_tree_reduce_ignore(self):
        """test the tree_reduce respects ignore"""
        tree = ["1", "2", ["3", "4"], [["5"], "6", "7", [[["8"]]]]]
        ignore = ["2", "4", "7"]
        flat = ["1", "3", "5", "6", "8"]
        self.assertEqual(
            shared.tree_reduce(tree, glue=".", ignore=ignore), flat
        )

    def test_3_tree_reduce_real(self):
        """some gets that failed once upon a time"""
        tree = ["", "validators"]
        ignore = ["", ""]
        flat = ["validators"]
        self.assertEqual(
            shared.tree_reduce(tree, glue=".", ignore=ignore), flat
        )

        tree = [["backend.outputs", "mke_cluster"], "validators"]
        ignore = ["", ""]
        flat = ["backend", "outputs", "mke_cluster", "validators"]
        self.assertEqual(
            shared.tree_reduce(tree, glue=".", ignore=ignore), flat
        )

    def test_4_tree_get(self):
        """does tree_get() work with all sorts of messy keys"""

        tree = {"1": {"2": {"3": {"4": "my value"}}}}

        self.assertEqual(shared.tree_get(tree, "1.2.3"), {"4": "my value"})
        self.assertEqual(shared.tree_get(tree, "1.2.3.4"), "my value")
        self.assertEqual(shared.tree_get(tree, "1.2.3.4."), "my value")
        self.assertEqual(shared.tree_get(tree, "1.2.3.4"), "my value")
        self.assertEqual(
            shared.tree_get(tree, ["1", "2", "3", "4"]), "my value"
        )
        self.assertEqual(shared.tree_get(tree, ["1.2", "3.4"]), "my value")
        self.assertEqual(
            shared.tree_get(tree, ["1", ["2", "3.4"]]), "my value"
        )

        self.assertEqual(
            shared.tree_get(tree, ["1", "2", ".", "", "3.", ".4"]), "my value"
        )
        self.assertEqual(
            shared.tree_get(tree, [[".", ""], "1.2.3.4", "..", [".", ["."]]]),
            "my value",
        )

    def test_5_tree_merge(self):

        source = {
            "1": "source 1",
            "2": "source 2",
            "4": "source 4",
            "5": {"1": "source 5.1"},
        }
        destination = {
            "3": {
                "1": "destination 3.1",
                "2": {"1": "destination 3.2.1", "2": "destination 3.2.2"},
            },
            "4": {"1": "third 4.1"},
            "5": "third 5",
        }

        merged = shared.tree_merge(source, destination)

        self.assertIsNotNone(merged)
        self.assertEqual(shared.tree_get(merged, "3.1"), "destination 3.1")
        self.assertEqual(shared.tree_get(merged, "4"), "source 4")
        self.assertEqual(shared.tree_get(merged, "5.1"), "source 5.1")

        third = {}

    def test_get_listindexes(self):
        """can get reach into lists using numberic indexes"""

        test_list = [
            "1.2 0",
            "1.2 1",
            "1.2 2",
        ]
        wrapper_dict = {"2": test_list}
        tree = {"1": wrapper_dict}

        self.assertEqual(shared.tree_get(tree, "1"), wrapper_dict)
        self.assertEqual(shared.tree_get(tree, "1.2"), test_list)
        self.assertEqual(shared.tree_get(tree, "1.2.0"), test_list[0])
        self.assertEqual(shared.tree_get(tree, "1.2.2"), test_list[2])

        with self.assertRaises(IndexError):
            shared.tree_get(tree, "1.2.5")

        with self.assertRaises(ValueError):
            shared.tree_get(tree, "1.2.string")



"""

Config formatting code.

@NOTE this code has gone through three major iterations, starting with a single
  function, expanded to be module, expanded further with regex, expanded again
  with centralized processing using regex, and finally this parser.  Some
  of the comments may be out of date.

"""
import logging
from typing import Any, Dict, List
from enum import Enum

from .plugin import Type

logger = logging.getLogger("configerus.format")

DEFAULT_FORMATTER_PLUGIN_ID = "get"
"""This plugin is used for formatting when a target doesn't specify plugin."""
FORMATTER_PLUGIN_ID_PASSTHROUGH = "value"
"""This config formatter plugin id just passes the target through."""
FORMATTER_PLUGIN_ID_INTEPRET = "special"
"""This config formatter plugin id to intepret the value."""

FORMATTER_SPECIAL_VALUES: Dict[str, Any] = {
    "Null": None,
    "None": None,
    "{}": {},
    "[]": [],
    "empty": "",
}
""" A map of interpation values for the interpet target """


class ParserState(Enum):
    """What the parser currently thinks it is looking at."""

    OUTSIDE = 0  # we are outside of a matching pattern
    START = 1  # we are inside a matching pattern, either a plugin or a target
    KEY = 2  # we know we are in a target key name
    IGNORE = 3  # we are ignoring the current context


class ParserStackNode:
    """Node in the stack.

    We parse a string using a language stack.
    """

    def __init__(self, parent, state=ParserState.OUTSIDE):
        """Initialize the language stack node.

        Parameters:
        -----------
        parent (node) : a stack needs to know its parent for popping
        state (State) : every node corresponds to a single state
        """
        self.parent = parent
        self.state = state

        self.plugin = None
        self.key = None
        """ some nodes track information collected for interpretation """

        self.parts = []
        """ list of subject parts that are a part of this node """

    def push(self, state):
        """Push a state child and return it."""
        return ParserStackNode(parent=self, state=state)

    def pop(self):
        """Get the node parent."""
        return self.parent

    def __str__(self):
        """Write to a string for debugging."""
        return "{}:{}[{}][p:{},k:{}]".format(
            self.parent, self.state.value, self.join(), self.plugin, self.key
        )

    def join(self):
        """Join the parts back together.

        Returns:
        --------
        If only 1 item is in the node, then it is returned raw, otherwise we
        return a string joing of all of the parts.

        @NOTE this can result in false identities when comparing dicts as
            strings to things like json, as Python uses single quotes, but
            JSON uses double.

        """
        if len(self.parts) == 1:
            return self.parts[0]

        return "".join(["{}".format(part) for part in self.parts])

    def append(self, part):
        """Append a part to the node."""
        self.parts.append(part)


class Formatter:
    """Object which replaces found syntax matches using formatting plugins."""

    START_MATCH = "{{"
    STOP_MATCH = "}}"
    PLUGINEND_MATCH = "::"
    DEFAULT_MATCH = "?"

    def __init__(
        self,
        config,
        default_plugin_for_target: str = DEFAULT_FORMATTER_PLUGIN_ID,
        default_plugin_for_default: str = FORMATTER_PLUGIN_ID_PASSTHROUGH,
        fail_on: List[str] = None,
    ):
        """Initialize the formatter."""
        self.config = config
        self.default_plugin_for_target = default_plugin_for_target
        self.default_plugin_for_default = default_plugin_for_default

        self.fail_on = fail_on if fail_on is not None else []
        """ For debugging, always fail on these values """

        self.special_values: Dict[str, Any] = FORMATTER_SPECIAL_VALUES
        """Special formatter key values."""

    def format(self, data: Any, default_label: str):
        """Perform a deep format."""
        return self.recursive_format(data=data, default_label=default_label)

    def recursive_format(self, data: Any, default_label: str):
        """Perform recursive deep formatting."""
        # try to iterate across the target
        # There is probably a more `python` approach that would cover more
        # iterables, as long as we can re-assign
        if isinstance(data, list):
            for index, value in enumerate(data):
                data[index] = self.recursive_format(
                    data=value, default_label=default_label
                )
        if isinstance(data, dict):
            for index, value in data.items():
                data[index] = self.recursive_format(
                    data=value, default_label=default_label
                )

        # strings get some searching for format actions
        elif isinstance(data, str):
            # if the entire target is the match, then replace whatever type
            # we get out of the config .get() call
            data = self.format_string(
                subject=data, default_label=default_label
            )

        else:
            # all sorts of primitives and custom objects are just ignored.
            pass

        return data

    # Not sure how to reduce branching here. This is the primary stack
    # processing method.
    # pylint: disable=too-many-branches
    def format_string(self, subject: str, default_label: str):
        """Format subject string.

        Looks for and processes any formatting tags in the subject.
        """
        # fast stop test.  if no start tags are in the string then just abort
        if self.START_MATCH not in subject:
            return subject

        subject_tokenized = subject
        for token in [
            self.START_MATCH,
            self.STOP_MATCH,
            self.PLUGINEND_MATCH,
            self.DEFAULT_MATCH,
        ]:
            subject_tokenized = subject_tokenized.replace(
                token, "@@@{}@@@".format(token)
            )
        # subject string split/exploded into words
        subject_parts: List[str] = subject_tokenized.split("@@@")

        stack = ParserStackNode(parent=None, state=ParserState.OUTSIDE)
        # print ("PRE: {} => {}".format(subject, subject_parts))

        for part in subject_parts:

            if part == "":
                # print("SKIP EMPTY -> {}".format(part))
                continue

            if part in [self.START_MATCH]:
                stack = stack.push(ParserState.START)
                stack.plugin = self.default_plugin_for_target
                # print("START: '{}' :: {}".format(part, stack))

            elif part in [self.STOP_MATCH]:

                if stack.state is ParserState.START:
                    # we had started a tag, we had collected only a tag key
                    # and now we see s stop tag, so assume that is all we get
                    key = stack.join()
                    stack = stack.pop()
                    plugin = (
                        stack.plugin
                        if stack.plugin is not None
                        else self.default_plugin_for_target
                    )

                    replace = self.find_replacement(
                        plugin, key, default_label=default_label
                    )
                    stack.append(replace)
                    # print("STOP: (start) '{}' :: {}".format(part, stack))

                elif stack.state is ParserState.KEY:
                    # we had started a tag, and had collected
                    key = stack.join()
                    stack = stack.pop()
                    plugin = (
                        stack.plugin
                        if stack.plugin is not None
                        else self.default_plugin_for_target
                    )

                    replace = self.find_replacement(
                        plugin, key, default_label=default_label
                    )
                    stack.append(replace)
                    # print("STOP: (key) '{}' :: {}".format(part, stack))

                elif stack.state is ParserState.IGNORE:
                    # we are ignoring this stuff
                    stack = stack.pop()
                    # print("STOP: (ignore) '{}' :: {}".format(part, stack))

                else:
                    # print("STOP: (what's this)'{}'::{}".format(part, stack))
                    raise ValueError("Unexpected tag end : {}".format(stack))

            elif stack.state in [ParserState.IGNORE]:
                # If we are in an ignore state, only track tag start and stop
                # so skip everything below
                # print("IGNORING: {}".format(part))
                pass

            elif part in [self.PLUGINEND_MATCH]:
                plugin = stack.join()
                stack = stack.pop()

                stack.plugin = plugin
                stack = stack.push(ParserState.KEY)
                # print("PLUGIN: '{}' :: {}".format(part, stack))

            elif part in [self.DEFAULT_MATCH]:
                key = stack.join()
                stack = stack.pop()

                stack.key = key
                plugin = (
                    stack.plugin
                    if stack.plugin is not None
                    else self.default_plugin_for_target
                )

                try:
                    replace = self.find_replacement(
                        plugin, key, default_label=default_label
                    )
                    stack.append(replace)

                    # if the replacement retrieval did not throw an exception
                    # then ignore the rest of the tag
                    stack = stack.push(ParserState.IGNORE)
                    # print(f"DEFAULT:IGNORING'{part}'::{stack}")
                except KeyError:
                    # if retrieving a replacement raised an exception then we
                    # switch into DEFAULT mode
                    stack.plugin = self.default_plugin_for_default
                    stack = stack.push(ParserState.START)
                    # print(f"DEFAULT: PROCESSING DEFAULT '{part}' :: {stack}")

            else:
                stack.append(part)
                # print("INC: {}".format(part))

        # print("DONE:  {}".format(stack))
        return stack.join()

    def find_replacement(self, plugin: str, key: str, default_label: str):
        """Use a plugin and key to find a replacement."""
        if plugin == FORMATTER_PLUGIN_ID_PASSTHROUGH:
            # In this case we are just passing through.  This is used
            # primarily by the default processor as it has little value from
            # real source.
            return key

        if plugin == FORMATTER_PLUGIN_ID_INTEPRET:
            # In this case interpret the value as something special
            return self.interpret(key)

        if self.config.plugins.has_plugin(
            instance_id=plugin, type=Type.FORMATTER
        ):
            plugin = self.config.plugins.get_plugin(
                instance_id=plugin, type=Type.FORMATTER
            )
        elif self.config.plugins.has_plugin(
            plugin_id=plugin, type=Type.FORMATTER
        ):
            plugin = self.config.plugins.get_plugin(
                plugin_id=plugin, type=Type.FORMATTER
            )
        else:
            logger.error(
                [
                    instance.instance_id
                    for instance in self.config.plugins.instances
                ]
            )
            raise RuntimeError(
                f"Unknown format plugin '{Type.FORMATTER.value}::{plugin}', "
                "no such plugin has been added."
            )

        return plugin.format(key=key, default_label=default_label)

    def interpret(self, key: str):
        """Interpret that value to mean something special."""
        try:
            return self.special_values[key.strip()]
        except Exception as err:
            raise ValueError(
                "Could not interpret special key '{}'".format(key)
            ) from err



"""

Test config's copy()

Here we test that the Config object is able to make copies of itself, and that
changing those copies doesn't affect originals.

"""

from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
import configerus
import unittest
import logging

logging.basicConfig()
logger = logging.getLogger("config_copy")
logger.setLevel(level=logging.INFO)


""" TESTS """


class ConfigBehaviour(unittest.TestCase):
    def test_copy_safety(self):
        """Test that config copy allows overloads and doesn't modify the source"""

        logger.debug("Building empty config object")
        config = configerus.new_config()

        config.add_source(PLUGIN_ID_SOURCE_DICT, "orig", 80).set_data(
            {"copy": {"one": "orig 1"}}
        )

        config_copy_orig = config.load("copy")

        copy1 = config.copy()
        copy1.add_source(PLUGIN_ID_SOURCE_DICT, "copy1", 81).set_data(
            {"copy": {"one": "copy1 1", "two": "copy1 2"}}
        )
        copy2 = config.copy()
        copy2.add_source(PLUGIN_ID_SOURCE_DICT, "copy2", 82).set_data(
            {"copy": {"one": "copy2 1", "two": "copy2 2"}}
        )

        config_copy_late = config.load("copy")
        config1_copy = copy1.load("copy")
        config2_copy = copy2.load("copy")

        logger.debug("LOADED: AFTER: orig: %s", config_copy_orig.data)
        logger.debug("LOADED: AFTER: 1   : %s", config1_copy.data)
        logger.debug("LOADED: AFTER: 2   : %s", config2_copy.data)
        logger.debug("LOADED: AFTER: late: %s", config_copy_late.data)

        # check original values
        self.assertEqual(config_copy_orig.get("one"), "orig 1")
        with self.assertRaises(KeyError):
            config_copy_orig.get("two")
        # check that copied config didn't modify original
        self.assertEqual(
            config_copy_orig.get("one"), config_copy_late.get("one")
        )
        with self.assertRaises(KeyError):
            config_copy_late.get("two")

        self.assertEqual(config1_copy.get("one"), "copy1 1")
        self.assertEqual(config1_copy.get("two"), "copy1 2")
        self.assertEqual(config2_copy.get("one"), "copy2 1")
        self.assertEqual(config2_copy.get("two"), "copy2 2")
